#!/usr/bin/env ruby

require 'ftools'

if ARGV.length < 1
  puts 'missing DIRECTORY argument'
  abort
end
PATCH_ROOT = ARGV[0] + '/'

# COPY
['mm/kmemprof.c','include/linux/kmemprof.h'].each do |i|
  File.copy(File.basename(i), PATCH_ROOT+ i)
end

# REWRITE
def array_insert (lines, i, l)
  lines.slice(0..i) + l + lines.slice((i+1)..-1)
end

def strarray_insert_at(lines, re, l, off=0)
  index = -1
  lines.each_index do |i|
    if lines[i] =~ re
      index = i
      break
    end
  end
  abort if index < 0
  array_insert(lines, index + off, l)
end

patch = {}
patch['Makefile'] = Proc.new do |lines|
  lines.collect do |l|
    l = l.chomp + ' -kmemprof' if l =~ /^EXTRAVERSION =/
    l
  end
end

patch['include/linux/slab.h'] = Proc.new do |lines|
  strarray_insert_at(lines, /^extern unsigned int kmem_cache_size.*;/,
    ["extern const char * kmem_cache_name(kmem_cache_t *cachep);\n"])
end

patch['kernel/ksyms.c'] = Proc.new do |lines|
  strarray_insert_at(lines, /^EXPORT_SYMBOL\(kmem_cache_size\);/, 
    ["EXPORT_SYMBOL(kmem_cache_name);\n"])
end

patch['mm/Makefile'] = Proc.new do |lines|
  tmp = lines.collect do |l|
    l = l.sub(/\\/, ' kmemprof.o \\') if l =~ /^obj-y\s+:=.*\\/
    l
  end
end

patch['mm/slab.c'] = Proc.new do |lines|

  p 'inc'
  lines = strarray_insert_at(lines, /^#include[ \t]*<linux\/slab\.h>/,
    ["#include\t<linux/kmemprof.h>\n"])

  p 'cache'

  cache_alloc = <<EOF
{
static inline
void * wrapped__cache_alloc (kmem_cache_t *cachep, int flags);

	void * ret;
	struct kmemprof_event event;

	ret = wrapped__cache_alloc(cachep, flags);

	event.type = KMEM_ALLOC;
	event.cache = cachep;
	event.flags = flags;
	event.obj = ret;
	event.frame = __builtin_frame_address(0);
	event.eip = __builtin_return_address(0);
	kmemprof_notifier_call_chain(KMEM_ALLOC, &event);

	return ret;
}

static inline void * wrapped__cache_alloc (kmem_cache_t *cachep, int flags)
EOF

  lines = strarray_insert_at(lines, /^static[ \t]*.*__cache_alloc[ \t]*\(kmem_cache_t[ \t]*\*.*,[ \t]*int[ \t]+.*\)$/,
  cache_alloc.split(/\n/).collect { |x| x + "\n" })

  cache_free = <<EOF
{
static inline
void wrapped__cache_free (kmem_cache_t *cachep, void* objp);

	struct kmemprof_event event;

	event.type = KMEM_FREE;
	event.cache = cachep;
	event.flags = 0;
	event.obj = objp;
	event.frame = __builtin_frame_address(0);
	event.eip = __builtin_return_address(0);
	kmemprof_notifier_call_chain(KMEM_FREE, &event);

	wrapped__cache_free(cachep, objp);

	return;
}

static inline void wrapped__cache_free (kmem_cache_t *cachep, void* objp)
EOF

    p 'free'
    lines = strarray_insert_at(lines, /^static[ \t]+.*__cache_free[ \t]*\(kmem_cache_t[ \t]+\*.*,[ \t]*void[ \t]*\*.*\)$/,
    cache_free.split(/\n/).collect { |x| x + "\n" })

    cache_name = <<EOF
const char * kmem_cache_name(kmem_cache_t *cachep)
{
	        return cachep->name;
}

EOF
    p 'name'
    lines = strarray_insert_at(lines, /^unsigned int kmem_cache_size[ \t]*\(kmem_cache_t[ \t]*\*[ \t]*.*\)$/,
    cache_name.split(/\n/).collect { |x| x + "\n" }, -1)

end

patch.each_pair do |f,handler|
  f = PATCH_ROOT + f
  lines = IO.readlines(f)
  lines = handler.call(lines)
  File.open(f, 'w') { |file| lines.each {|l| file.write(l)} }
end

